#!/usr/bin/python
#
# Perform calculations on USGS National Water Information System rdb files
# http://waterdata.usgs.gov/nwis/?tab_delimited_format_info
#
# Copyright 2010 Alex Chiang <alex@chizang.net>
#
# This program is distributed under the terms of the
# GNU General Public License version 3.

import os
import re
import sys
from optparse import OptionParser

class Report(object):
    def __init__(self, gage_file):
        self.years = {}

        def parse_year(date):
            '''Expect dates to be in YYYY-MM-DD format.
            We return the water year, not calendar year here.'''
            (year, month, day) = date.split('-')
            if int(month) < 10:
                return year
            else:
                return str(int(year) + 1)

        def parse_flow(flow):
            '''Flow data may have qualification codes, delimited by an
            underscore. If so, ignore the qual code and just grab the
            numerical value. Coerce all values to floating point numbers
            for consistent calculations later.'''
            if re.search('_', flow):
                return float(flow.split('_')[0])
            if flow == '':
                return float(0)
            return float(flow)

        gf = open(gage_file, 'r')
        for line in gf:
            # All data lines begin with 'USGS'
            if not re.match('USGS', line):
                continue

            # Thanks to USGS, all data lines are tab-delimited. Lines without
            # flow data will still contain tabs, delimiting empty columns.
            # Thus, if we split on \t, we are guaranteed that we will always
            # have a field for the flow data, even if it is just the empty
            # value of ''.
            fields = line.split('\t')
            year = parse_year(fields[2])
            flow = parse_flow(fields[3])

            # If this is the first time we've seen this year, add it to our
            # global dictionary of all years. We initialize it with an
            # empty list.
            if not year in self.years:
                self.years[year] = []

            # Append each day's flow data to the year. When the loop completes,
            # we end up with a dictionary that conceptually is:
            #    a) indexed by year
            #    b) each year has an attached list of each day's flow data
            # and it looks like:
            #    year1: [day1, day2, day3, ..., day365]
            #    year2: [day1, day2, day3, ..., day365]
            self.years[year].append(flow)
        gf.close()

    def average_daily_flow(self):
        print "Average Daily Flow"
        print "------------------"
        print "Year\tFlow"
        for year, days in sorted(self.years.items()):
            sum = 0.0
            for d in days:
                sum = sum + d
            avg = sum / len(days)

            # Print with 2 decimal places of precision. Change as desired.
            print "%s\t%.2f" % (year, avg)

    def annual_minimum_flow(self):
        print "Annual Minimum Flow"
        print "-------------------"

        # This is a little sneaky. First, define an impossible flow value.
        # sys.maxint is 9223372036854775807 on my machine.
        IMPOSSIBLE = sys.maxint
        for year, days in sorted(self.years.items()):
            # Now, for each year, we initialize the min_flow to the impossible
            # value. We know that we're going to find values less than this
            # value later on...
            min_flow = IMPOSSIBLE
            for d in days:
                # Finally, if the day's flow is less than our previous min_flow
                # (which started off as IMPOSSIBLE), declare this day's flow
                # as our new min... unless the flow data is 0. We do this
                # to guard against days where we did not record flow data.
                # No flow data is not the same thing as minimal flow. So we
                # require some non-zero value before we consider this day's
                # flow to be our minimum flow.
                if d < min_flow and d > 0:
                    min_flow = d
            # If we get through all the days in the year and our min_flow is
            # still IMPOSSIBLE, it means that every single day had a 0.0
            # flow value. This describes the scenario where we didn't collect
            # any data for the entire year. In this case, just give up and
            # describe the minimum flow for the entire year as 0.0, and later
            # when we look at the output, we'll just have to know that 0.0
            # means we didn't collect any data for the year.
            if min_flow == IMPOSSIBLE:
                min_flow = 0
            print "%s\t%.2f" % (year, min_flow)

    def zero_flow_days(self):
        print "Zero Flow Days"
        print "--------------"
        for year, days in sorted(self.years.items()):
            zeroes = days.count(0.0)
            # Do not display years with either no zero flow days, and do not
            # display years in which every single day is zero flow (meaning
            # we did not collect data for that year). Of course we use the
            # greater than or equal to 365 condition to catch leap years.
            if zeroes == 0 or zeroes >= 365:
                continue
            # Here, the %3d means 'right justify output with max width of 3
            # characters'.
            print "%s\t%3d" % (year, zeroes)

def main():
    usage = "usage: %prog -f <gage_file_name> [-c <calculation>]"
    parser = OptionParser(usage=usage)
    parser.add_option('-f', dest='gage_file', help='Specify the gage file')
    parser.add_option('-c', dest='calc',
        help='Choose a specific calculation to run. Leave unspecified to '
             'run all known calculations. Possible arguments are:\n'
             '\taverage_daily_flow'
             '\tannual_minimum_flow'
             '\tzero_flow_days')

    opts, args = parser.parse_args()
    if opts.gage_file == None:
        parser.error("gage file name required")
        sys.exit()

    # Parse the gage file, but don't do any calculations yet
    report = Report(opts.gage_file)

    # If no specific calculation is chosen, then run every calc we know about.
    if opts.calc == None:
        # The power of python introspection. First, look at all the attributes
        # of our Report object...
        for attr in dir(report):
            # Avoid any built-in functions and our __init__ function.
            if re.match('_', attr):
                continue
            # Finally, for any attributes left, if we can call it as a
            # function call, do it. This helps us avoid from manually
            # calling every function we define in the Report class. Even
            # if we define more calculations in the future, we won't have
            # to change this loop here. We will need to modify the help text
            # above though, where we describe the calculations available
            # to the user.
            if callable(getattr(report, attr)):
                getattr(report, attr)()
                print
    else:
        getattr(report, opts.calc)()

if __name__ == "__main__":
    main()
